//! Prymn messaging framework

use async_nats::{ConnectError, ConnectOptions, SubscribeError};
use thiserror::Error;
use tokio_stream::StreamExt;

pub use self::v1::Command;
pub use self::v1::OutgoingMessage;

#[derive(Debug, Error)]
pub enum MessagingError<'a> {
    #[error("failed to connect to message broker")]
    ConnectError(#[from] ConnectError),
    #[error("failed to listen for messages from broker")]
    SubscribeError(#[from] SubscribeError),
    #[error("unknown subject '{0}'")]
    UnknownSubject(&'a str),
    #[error("parse error on message format")]
    MessageFormatError(#[from] serde_json::Error),
}

mod v1 {
    use bytes::Bytes;
    use serde::Deserialize;

    use super::MessagingError;

    pub(super) const PREFIX: &'static str = "agents.v1";

    #[derive(Debug)]
    pub enum OutgoingMessage {
        Health(String),
    }

    impl OutgoingMessage {
        pub fn subject(&self, sender: &str) -> String {
            match self {
                OutgoingMessage::Health(_) => format!("{}.{}.health", PREFIX, sender),
            }
        }

        pub fn into_payload(self) -> Bytes {
            match self {
                OutgoingMessage::Health(data) => Bytes::from(data.into_bytes()),
            }
        }
    }

    #[derive(Debug)]
    pub enum Command {
        Exec(ExecArgs),
    }

    impl Command {
        pub fn from_message<'a>(
            subject: &'a str,
            payload: &'a [u8],
        ) -> Result<Self, MessagingError<'a>> {
            match subject {
                "exec" => Ok(Command::Exec(serde_json::from_slice(payload)?)),
                _ => Err(MessagingError::UnknownSubject(subject)),
            }
        }
    }

    /// An operating system program execution.
    #[derive(Debug, Deserialize)]
    pub(super) struct ExecArgs {
        user: String,
        program: String,
        #[serde(default)]
        args: Vec<String>,
    }

    // /// A system update message received when an OS update is requested
    // #[derive(Deserialize)]
    // struct SystemUpdate {
    //     /// `true` when a simulated update should occur
    //     simulate: bool,
    // }

    // /// An agent update message received when an agent binary update is requested
    // #[derive(Deserialize)]
    // struct AgentUpdate {
    //     /// `true` when a simulated update should occur
    //     simulate: bool,
    // }
}

#[derive(Debug, Clone)]
pub struct Client {
    id: String,
    client: async_nats::Client,
}

impl Client {
    #[tracing::instrument]
    pub async fn connect(id: &str) -> Result<Self, MessagingError> {
        let client = ConnectOptions::new()
            .name(format!("Prymn Agent {id}"))
            .custom_inbox_prefix(format!("_INBOX_{id}"))
            .user_and_password("demo_agent".to_owned(), "demo_agent_password".to_owned())
            .connect("localhost")
            .await?;

        tracing::debug!("connected to NATS");

        let prefix = format!("{}.{}.*", v1::PREFIX, id);
        let mut routing_subscriber = client.subscribe(prefix.clone()).await?;

        tokio::spawn(async move {
            tracing::debug!("v1 command routing subscribed on {}", &prefix);

            while let Some(data) = routing_subscriber.next().await {
                let subject = data.subject.trim_start_matches(&prefix[..prefix.len() - 1]);

                match Command::from_message(subject, &data.payload) {
                    Ok(command) => {
                        tracing::debug!(?command);
                        // TODO: Think if we might want to queue commands
                        // handle_command(command);
                    }
                    Err(err) => {
                        tracing::warn!(cause = %err, "failed to parse received message");
                        tracing::debug!(payload = %String::from_utf8_lossy(&data.payload));
                    }
                }
            }
        });

        Ok(Self {
            id: id.to_owned(),
            client,
        })
    }

    #[tracing::instrument(skip(self))]
    pub async fn publish(&self, msg: OutgoingMessage) {
        let subj = msg.subject(&self.id);
        tracing::debug!(subject = &subj, "publish");
        if let Err(err) = self.client.publish(subj, msg.into_payload()).await {
            tracing::error!(cause = %err, "publish error");
        }
    }
}

// pub fn receive_requests(agent: Client) {
//     let prefix = format!("{}.{}.*", v1::PREFIX, id);
//     let mut routing_subscriber = client.subscribe(prefix.clone()).await?;
//     let subscriber = agent.subscribe("").await;
//     tracing::debug!("v1 command routing subscribed on {}", &prefix);
// }
//
// fn handle_command(command: Command) {
//     match command {
//         Command::Exec(_) => todo!(),
//     }
// }
